@@ -1,9 +1,73 @@
 # perform various data augmentations 
 
+
+## callbacks supported
+
+
+
+def rampup_lrfn(epoch):
+
+	"""
+	Fine Tune the pretrained model weights in a better and smooth manner.
+	Use this as a callback feature during the model training.
+	"""
+	
+	if epoch < LR_RAMPUP_EPOCHS:
+		lr = (LR_MAX - LR_START) / LR_RAMPUP_EPOCHS * epoch + LR_START
+	elif epoch < LR_RAMPUP_EPOCHS + LR_SUSTAIN_EPOCHS : 
+		lr = LR_MAX
+	else:
+		lr = (LR_MAX - LR_MIN) + LR_EXP_DECAY**(epoch - LR_RAMPUP_EPOCHS - LR_SUSTAIN_EPOCHS) + LR_MIN
+
+	return lr
+
+
+
+def stepped_lrfn(epoch):
+	return 1e-4 * (0.75 ** np.floor(epoch / 2))
+
+def step_decay(epoch):
+	initAlpha = 0.01
+	factor = 0.25
+	dropEvery = 5
+	
+	alpha = initAlpha * (factor ** np.floor((1+epoch) / dropEvery))
+	return float(alpha)
+
+def simple_lrfn(epoch):
+
+	if epoch < 6:
+		return 0.001
+	elif 6 < epoch < 12 : 
+		return 0.0005
+	elif 12 < epoch < 18 : 
+		return 0.00008
+	else:
+		return 0.000006
+
+	return
+
+
+
+def define_callbacks(lr_scheduler = None):
+	
+	global callbacks
+	
+	callbacks = {}
+	if lr_scheduler is not None:
+		callbacks['lr_scheduler'] = lr_scheduler
+	else:
+		callbacks['lr_scheduler'] = None
+
+	if len(list(callbacks.keys())) == 0:
+		print("Error! No callbacks defined but callback called")
+
+
 def define_augmentations(flip_left_right = False, hue = None, contrast = None, brightness = None):
 	
 	
 	global augments
+	## add random saturation, random crop, zoom, shearing, rotation etc
 	
 	augments = {}
 
@@ -47,7 +111,7 @@ def augmentations(image, label):
 
 
 
-def augment_and_training_report(models_list, GCS_DS_PATH, train_tfrec_path, val_tfrec_path, batch_size, tpu, n_class, steps_per_epoch, epochs, classification_model = 'default', freeze = False, input_shape = [512,512,3], activation = 'softmax', weights = "imagenet", optimizer = "adam", loss = "sparse_categorical_crossentropy", metrics = "sparse_categorical_accuracy", callbacks = None, plot = False):
+def augment_and_training_report(models_list, GCS_DS_PATH, train_tfrec_path, val_tfrec_path, batch_size, tpu, n_class, steps_per_epoch, epochs, classification_model = 'default', freeze = False, input_shape = [512,512,3], activation = 'softmax', weights = "imagenet", optimizer = "adam", loss = "sparse_categorical_crossentropy", metrics = "sparse_categorical_accuracy", plot = False):
 
 
 	import pandas as pd
@@ -98,24 +162,64 @@ def augment_and_training_report(models_list, GCS_DS_PATH, train_tfrec_path, val_
 
 
 
-def augment_and_train(model, GCS_DS_PATH, train_tfrec_path, val_tfrec_path, batch_size, epochs, steps_per_epoch, callbacks = None, plot = False):
+
+
+
+def set_learning_rate(model, epoch):
+	
+	global callbacks
+	
+	if callbacks['lr_scheduler'] is not None:
+		lr_scheduler = callbacks['lr_scheduler']
+		if lr_scheduler == 'rampup':
+			lr = rampup_lrfn(epoch)
+			K.set_value(model.optimizer.learning_rate, lr)
+			print(f"Learning Rate changed to {K.eval(model.optimizer.lr)}")
+		elif lr_scheduler == 'simple':
+			lr = simple_lrfn(epoch)
+			K.set_value(model.optimizer.learning_rate, lr)
+			print(f"Learning Rate changed to {K.eval(model.optimizer.lr)}")
+		elif lr_scheduler == 'step_decay':
+			lr = step_decay(epoch)
+			K.set_value(model.optimizer.learning_rate, lr)
+			print(f"Learning Rate changed to {K.eval(model.optimizer.lr)}")
+		elif lr_scheduler == 'stepped':
+			lr = step_decay(epoch)
+			K.set_value(model.optimizer.learning_rate, lr)
+			print(f"Learning Rate changed to {K.eval(model.optimizer.lr)}")
+		
+	
+
+
+def augment_and_train(model, GCS_DS_PATH, train_tfrec_path, val_tfrec_path, batch_size, epochs, steps_per_epoch, plot = False):
 
 	from quick_ml.begin_tpu import get_training_dataset
 	from quick_ml.begin_tpu import get_validation_dataset
 
 	overall = {}
 	
+	
+	
 	for epoch in range(epochs):
 	
 		if epoch == 0:
 			print(f"\nEpoch # {epoch + 1}\n")
-			history = model.fit(get_training_dataset(GCS_DS_PATH, train_tfrec_path, batch_size, augmentation = True), validation_data = get_validation_dataset(GCS_DS_PATH, val_tfrec_path, batch_size, augmentation = True), epochs = 1, steps_per_epoch = steps_per_epoch, batch_size = batch_size, verbose = 1, callbacks = callbacks)
+			print(f"\nCurrent Learning Rate -> {K.eval(model.optimizer.lr}")
+			history = model.fit(get_training_dataset(GCS_DS_PATH, train_tfrec_path, batch_size, augmentation = True), validation_data = get_validation_dataset(GCS_DS_PATH, val_tfrec_path, batch_size, augmentation = True), epochs = 1, steps_per_epoch = steps_per_epoch, batch_size = batch_size, verbose = 1)
+			if callbacks is not None:
+				set_callbacks(model, epoch + 1, callbacks)
 			
 			for k,v in history.history.items():
 				overall[k] = v
+				
+				
 		else:
+		
 			print(f"\nEpoch # {epoch + 1}\n")
-			history = model.fit(get_training_dataset(GCS_DS_PATH, train_tfrec_path, batch_size, augmentation = True), validation_data = get_validation_dataset(GCS_DS_PATH, val_tfrec_path, batch_size, augmentation = True), epochs = 1, steps_per_epoch = steps_per_epoch, batch_size = batch_size, verbose = 1, callbacks = callbacks)
+			print(f"\nCurrent Learning Rate -> {K.eval(model.optimizer.lr}")
+			history = model.fit(get_training_dataset(GCS_DS_PATH, train_tfrec_path, batch_size, augmentation = True), validation_data = get_validation_dataset(GCS_DS_PATH, val_tfrec_path, batch_size, augmentation = True), epochs = 1, steps_per_epoch = steps_per_epoch, batch_size = batch_size, verbose = 1)
+			if callbacks is not None:
+				set_callbacks(model, epoch + 1, callbacks)
 			
 			for k,v in history.history.items():
 				overall[k].extend(v)
@@ -129,8 +233,8 @@ def augment_and_train(model, GCS_DS_PATH, train_tfrec_path, val_tfrec_path, batc
 		
 		
 
-		sns.lineplot(x = list(range(1, epochs + 1)), y = overall[keys[0]], label = 'Training Accuracy');
-		sns.lineplot(x = list(range(1, epochs + 1)), y = overall[keys[2]], label = 'Validation Accuracy');
+		sns.lineplot(x = list(range(1, epochs + 1)), y = overall[keys[0]], label = 'Training Accuracy', color = 'k');
+		sns.lineplot(x = list(range(1, epochs + 1)), y = overall[keys[2]], label = 'Validation Accuracy', color = 'red');
 		plt.show()
 
 
@@ -142,8 +246,25 @@ def augment_and_train(model, GCS_DS_PATH, train_tfrec_path, val_tfrec_path, batc
 if __name__ == "__main__":
 	pass
 else:
+
+
+	global LR_MAX
+	global LR_MIN
+	global LR_RAMPUP_EPOCHS
+	global LR_SUSTAIN_EPOCHS
+	global LR_EXP_DEC
+	global LR_START
+	
 	import tensorflow as tf
 	if tf.__version__ == '2.2.0':
 		pass
 	else:
-		print("Error! Tensorflow Version mismatch. Please refer to the documentation to know the supported version of tensorflow.")
\ No newline at end of file
+		print("Error! Tensorflow version mismatch...")
+	import numpy as np
+	LR_MAX = 0.00005 * 8
+	LR_MIN = 0.00001
+	LR_RAMPUP_EPOCHS = 5
+	LR_SUSTAIN_EPOCHS = 0
+	LR_EXP_DECAY = 0.8
+	LR_START = 0.00001
+	import tensorflow.keras.backend as K
\ No newline at end of file
